////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// FFGL.cpp
//
// FreeFrame is an open-source cross-platform real-time video effects plugin system.
// It provides a framework for developing video effects plugins and hosts on Windows,
// Linux and Mac OSX.
//
// FreeFrameGL (FFGL) is an extension to the FreeFrame spec to support video processing
// with OpenGL on Windows, Linux, and Mac OSX.
//
// Copyright (c) 2018 www.freeframe.org
// All rights reserved.
//
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Redistribution and use in source and binary forms, with or without modification,
//	are permitted provided that the following conditions are met:
//
//  * Redistributions of source code must retain the above copyright
//    notice, this list of conditions and the following disclaimer.
//  * Redistributions in binary form must reproduce the above copyright
//    notice, this list of conditions and the following disclaimer in
//    the documentation and/or other materials provided with the
//    distribution.
//  * Neither the name of FreeFrame nor the names of its
//    contributors may be used to endorse or promote products derived
//    from this software without specific prior written permission.
//
//
//	THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
//	ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
//	WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
//	IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
//	INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
//	BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
//	DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
//	OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
//	OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
//	OF THE POSSIBILITY OF SUCH DAMAGE.
//
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// First version, Marcus Clements (marcus@freeframe.org)
// www.freeframe.org
//
// FreeFrame 1.0 upgrade by Russell Blakeborough
// email: boblists@brightonart.org
//
// FreeFrame 1.0 - 03 upgrade
// and implementation of FreeFrame SDK methods by Gualtiero Volpe
// email: Gualtiero.Volpe@poste.it
//
// FFGL upgrade by Trey Harrison
// email: trey@harrisondigitalmedia.com
//
// FFGL 2.0 by Menno Vink (menno@resolume.com)
// www.resolume.com
// -Removed the old FreeFrame functionality, and thus CPU processing is no longer supported.
// -Completed the FF_TYPE_OPTION parameter type. Plugins can add parameter elements which hosts
//  should use to show a dropdown styled parameter. When one of the options is selected the parameter
//  will get a value equal to that of the option's value.
// -Completed the FF_TYPE_BUFFER parameter type. When this parameter is used it should provide a usage indication
//  which the host can then use to fill that buffer with the requested data. This can be used by plugins to
//  access the host's global fft data for example.
//
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Includes
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

#include "PluginBase.h"
#include <memory.h>
#include "../glsdk_0_5_2/glload/include/gl_load.h"

namespace ffgl
{

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Static and extern variables used in the FreeFrame SDK
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

extern PluginInfo* g_CurrPluginInfo;

static PluginBase* s_pPrototype = NULL;

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// FreeFrame SDK default implementation of the FreeFrame global functions.
// Such function are called by the plugMain function, the only function a plugin exposes.
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

bool InitGLExts()
{
#if defined( WIN32 )
	static bool triedInit  = false;
	static bool initResult = false;
	if( triedInit )
		return initResult;
	triedInit = true;

	if( ogl_LoadFunctions() == ogl_LOAD_FAILED )
		return false;
	initResult = true;
	return initResult;
#else
	return true;
#endif
}

// 2019-07-11 ffgl@subpixels.com - changed return type from (void*) to (const PluginInfoStruct*)
const PluginInfoStruct* GetPluginInfo()
{
	return g_CurrPluginInfo->GetPluginInfo();
}

FFResult Initialise()
{
	if( g_CurrPluginInfo == NULL )
		return FF_FAIL;

	if( s_pPrototype == NULL )
	{
		// Get the instantiate function pointer
		FPCREATEINSTANCEGL* factoryMethod = g_CurrPluginInfo->GetFactoryMethod();

		// Call the instantiate function
		FFResult ret = factoryMethod( &s_pPrototype );

		// Make sure the instantiate call worked
		if( ( ret == FF_FAIL ) || ( s_pPrototype == NULL ) )
			return FF_FAIL;

		return FF_SUCCESS;
	}

	return FF_SUCCESS;
}

FFResult Deinitialise()
{
	if( s_pPrototype != NULL )
	{
		delete s_pPrototype;
		s_pPrototype = NULL;
	}
	return FF_SUCCESS;
}

FFUInt32 GetNumParams()
{
	if( s_pPrototype == NULL )
	{
		FFResult ret = Initialise();
		if( ret == FF_FAIL )
			return FF_FAIL;
	}

	return s_pPrototype->GetNumParams();
}

// 2019-07-11 ffgl@subpixels.com - changed return type from (char*) to (const char*)
const char* GetParamName( FFUInt32 index )
{
	if( s_pPrototype == NULL )
	{
		FFResult ret = Initialise();
		if( ret == FF_FAIL )
			return NULL;
	}

	return s_pPrototype->GetParamName( index );
}

FFMixed GetParamDefault( FFUInt32 index )
{
	FFMixed retval = { FF_FAIL };

	if( s_pPrototype == NULL )
	{
		FFResult ret = Initialise();
		if( ret == FF_FAIL )
			return retval;
	}

	return s_pPrototype->GetParamDefault( index );
}

// 2019-07-27 ffgl@subpixels.com - Use enum PluginCapability, rename index to capability
FFResult GetPluginCaps( PluginCapability capability )
{
	if( s_pPrototype == NULL )
	{
		FFResult ret = Initialise();
		if( ret == FF_FAIL )
			return FF_FAIL;
	}

	switch( capability )
	{
	case PluginCapability::FF_CAP_SETTIME:
		if( s_pPrototype->GetTimeSupported() )
			return FF_TRUE;
		else
			return FF_FALSE;
	case PluginCapability::FF_CAP_MINIMUMINPUTFRAMES:
		return s_pPrototype->GetMinInputs();
	case PluginCapability::FF_CAP_MAXIMUMINPUTFRAMES:
		return s_pPrototype->GetMaxInputs();
	}

	return FF_FALSE;
}

// 2019-07-11 ffgl@subpixels.com - changed return type from (void*) to (PluginExtendedInfoStruct*)
const PluginExtendedInfoStruct* GetExtendedInfo()
{
	return g_CurrPluginInfo->GetPluginExtendedInfo();
}

ParameterType GetParamType( FFUInt32 index )
{
	if( s_pPrototype == NULL )
	{
		FFResult ret = Initialise();
		if( ret == FF_FAIL )
			return ParameterType::FF_FAIL;
	}

	return s_pPrototype->GetParamType( index );
}

// 2019-07-11 ffgl@subpixels.com
// - changed return type from (void*) to (PluginBase*)
// - use const retval to return correct type
// - use FFMixed::StrValue and FloatValue instead of pointer recasting
// 2019-07-22 ffgl@subpixels.com
// - use enum ParameterType
PluginBase* InstantiateGL( const ViewportStruct* viewport )
{
	// Convenience constant to return on failure.
	PluginBase* const retval = (PluginBase*)FF_FAIL;

	if( g_CurrPluginInfo == NULL || viewport == NULL )
		return retval;

	// If the plugin is not initialized, initialize it
	if( s_pPrototype == NULL )
	{
		FFResult ret = Initialise();

		if( ( ret == FF_FAIL ) || ( s_pPrototype == NULL ) )
			return retval;
	}

	// Get the instantiate function pointer
	FPCREATEINSTANCEGL* factoryMethod = g_CurrPluginInfo->GetFactoryMethod();

	PluginBase* plugin = NULL;

	// Call the instantiate function
	FFResult ret = factoryMethod( &plugin );

	// Make sure the instantiate call worked
	if( ( ret == FF_FAIL ) || ( plugin == NULL ) )
		return retval;

	// Store self reference
	plugin->plugin = plugin;

	if( !InitGLExts() )
		return retval;

	// Call the InitGL method
	if( plugin->InitGL( viewport ) == FF_SUCCESS )
	{
		// Succes? we're done.
		return plugin;
	}

	// InitGL failed, delete the instance
	plugin->DeinitGL();
	delete plugin;

	return retval;
}

FFResult DeinstantiateGL( void* instanceID )
{
	PluginBase* p = (PluginBase*)instanceID;

	if( p != NULL )
	{
		// Disconnect if necessary
		if( p->isConnected )
		{
			p->Disconnect();
			p->isConnected = false;
		}

		p->DeinitGL();
		delete p;

		return FF_SUCCESS;
	}

	return FF_FAIL;
}

FFUInt32 GetNumParamElements( FFUInt32 index )
{
	if( s_pPrototype == NULL )
	{
		FFResult ret = Initialise();
		if( ret == FF_FAIL )
			return FF_FAIL;
	}

	return s_pPrototype->GetNumParamElements( index );
}

// 2019-07-11 ffgl@subpixels.com - changed return type from (char*) to (const char*)
const char* GetParamElementName( FFUInt32 paramIndex, FFUInt32 elementIndex )
{
	if( s_pPrototype == NULL )
	{
		FFResult ret = Initialise();
		if( ret == FF_FAIL )
			return NULL;
	}

	return s_pPrototype->GetParamElementName( paramIndex, elementIndex );
}

FFMixed GetParamElementDefault( FFUInt32 paramIndex, FFUInt32 elementIndex )
{
	FFMixed retval = { FF_FAIL };

	if( s_pPrototype == NULL )
	{
		FFResult ret = Initialise();
		if( ret == FF_FAIL )
			return retval;
	}

	return s_pPrototype->GetParamElementDefault( paramIndex, elementIndex );
}

// 2019-07-22 ffgl@subpixels.com - use enum ParameterUsage
ParameterUsage GetParamUsage( FFUInt32 index )
{
	if( s_pPrototype == NULL )
	{
		FFResult ret = Initialise();
		if( ret == FF_FAIL )
			return ParameterUsage::FF_FAIL;
	}

	return s_pPrototype->GetParamUsage( index );
}

const char* GetPluginShortName()
{
	if( s_pPrototype == NULL )
	{
		FFResult ret = Initialise();
		if( ret == FF_FAIL )
			return NULL;
	}

	const char* shortName = s_pPrototype->GetShortName();
	if( shortName == NULL )
		return NULL;

	return shortName;
}

// 2019-07-11 ffgl@subpixels.com
// - use FFMixed::GetParamRangeArgs instead of pointer recasting
FFMixed GetParamRange( FFMixed input )
{
	FFMixed retval = { FF_FAIL };

	if( s_pPrototype == NULL )
	{
		FFResult ret = Initialise();
		if( ret == FF_FAIL )
			return retval;
	}

	auto args = input.GetParamRangeArgs;

	RangeStruct range = s_pPrototype->GetParamRange( args->parameterNumber );

	// WARNING: args points to the host's structure
	// args->range is reserved for the output
	args->range = range;

	retval.UIntValue = FF_SUCCESS;
	return retval;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Implementation of plugMain, the one and only exposed function
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

// 2019-07-11 ffgl@subpixels.com
// - use new FFMixed union value types and pointers instead of pointer recasting
// 2019-07-22 ffgl@subpixels.com
// - use enums ParameterType, PluginCapability

#ifdef WIN32

FFMixed __stdcall plugMain( FFUInt32 functionCode, FFMixed inputValue, FFInstanceID instanceID )

#elif TARGET_OS_MAC

FFMixed plugMain( FFUInt32 functionCode, FFMixed inputValue, FFInstanceID instanceID )

#elif __linux__

FFMixed plugMain( FFUInt32 functionCode, FFMixed inputValue, FFInstanceID instanceID )

#endif

{
#if defined( _DEBUG ) && defined( WIN32 ) && defined( LAUNCH_DEBUGGER )
	LaunchDebugger();
#endif

	FFMixed retval = { FF_FAIL };

	// declare plugin - pointer to this instance
	PluginBase* plugin;

	// typecast DWORD into pointer to a PluginBase
	plugin = (PluginBase*)instanceID;

	switch( (FunctionCode) functionCode )
	{
	case FunctionCode::FF_GETINFO:
		retval.PluginInfo = GetPluginInfo();
		break;
	case FunctionCode::FF_INITIALISE_V2:
		retval.UIntValue = Initialise();
		break;
	case FunctionCode::FF_DEINITIALISE:
		retval.UIntValue = Deinitialise();
		break;
	case FunctionCode::FF_GETNUMPARAMETERS:
		retval.UIntValue = GetNumParams();
		break;
	case FunctionCode::FF_GETPARAMETERNAME:
		retval.StrValue = GetParamName( inputValue.UIntValue );
		break;
	case FunctionCode::FF_GETPARAMETERDEFAULT:
		retval = GetParamDefault( inputValue.UIntValue );
		break;
	case FunctionCode::FF_GETPARAMETERDISPLAY:
		if( plugin != NULL )
			retval.StrValue = plugin->GetParamDisplay( inputValue.UIntValue );
		else
			retval.StrValue = (char*)FF_FAIL;
		break;
	case FunctionCode::FF_SETPARAMETER:
		if( plugin != NULL )
		{
			// 2019-08-22 ffgl@subpixels.com - created SetMixedParameter function that takes a FFMixed value
			const auto args  = inputValue.SetParameterArgs;
			retval.UIntValue = plugin->SetMixedParam( args->ParameterNumber, args->NewParameterValue );
		}
		break;
	case FunctionCode::FF_GETPARAMETER:
		if( plugin != NULL )
		{
			// 2019-08-22 ffgl@subpixels.com - created GetMixedParameter function that returns a FFMixed value
			retval = plugin->GetMixedParam( inputValue.UIntValue );
		}
		break;
	case FunctionCode::FF_GETPLUGINCAPS:
		retval.UIntValue = GetPluginCaps( inputValue.PluginCapability );
		break;
	case FunctionCode::FF_GETEXTENDEDINFO:
		retval.PluginExtendedInfo = GetExtendedInfo();
		break;
	case FunctionCode::FF_GETPARAMETERTYPE:
		retval.ParameterType = GetParamType( inputValue.UIntValue );
		break;
	case FunctionCode::FF_GETINPUTSTATUS:
		if( plugin != NULL )
			retval.InputStatus = plugin->GetInputStatus( inputValue.UIntValue );
		break;
	case FunctionCode::FF_PROCESSOPENGL:
		if( plugin != NULL && inputValue.ProcessOpenGL != NULL )
		{
			// Ensure Connect has been called
			if( !plugin->isConnected )
			{
				plugin->Connect();
				plugin->isConnected = true;
			}

			retval.UIntValue = plugin->ProcessOpenGL( inputValue.ProcessOpenGL );
		}
		break;
	case FunctionCode::FF_INSTANTIATEGL:
		retval.FreeFrameGLPlugin = InstantiateGL( inputValue.Viewport );
		break;
	case FunctionCode::FF_DEINSTANTIATEGL:
		if( plugin != NULL )
			retval.UIntValue = DeinstantiateGL( plugin );
		break;
	case FunctionCode::FF_SETTIME:
		if( plugin != NULL && inputValue.DoubleValue != NULL )
			retval.UIntValue = plugin->SetTime( *inputValue.DoubleValue );
		break;
	case FunctionCode::FF_CONNECT:
		if( plugin != NULL )
		{
			retval.UIntValue = plugin->Connect();
			if( retval.UIntValue == FF_SUCCESS )
				plugin->isConnected = true;
		}
		break;
	case FunctionCode::FF_DISCONNECT:
		if( plugin != NULL )
		{
			retval.UIntValue = plugin->Disconnect();
			if( retval.UIntValue == FF_SUCCESS )
				plugin->isConnected = false;
		}
		break;
	case FunctionCode::FF_RESIZE:
		if( plugin != NULL )
			retval.UIntValue = plugin->Resize( inputValue.Viewport );
		break;
	case FunctionCode::FF_GETNUMPARAMETERELEMENTS:
		retval.UIntValue = GetNumParamElements( inputValue.UIntValue );
		break;
	case FunctionCode::FF_GET_PARAMETER_ELEMENT_NAME:
	{
		const auto args = inputValue.GetParamElementNameArgs;
		retval.StrValue = GetParamElementName( args->ParameterNumber, args->ElementNumber );
		break;
	}
	case FunctionCode::FF_GET_PARAMETER_ELEMENT_DEFAULT:
	{
		const auto args = inputValue.GetParamElementValueArgs;
		retval          = GetParamElementDefault( args->ParameterNumber, args->ElementNumber );
		break;
	}
	case FunctionCode::FF_SET_PARAMETER_ELEMENT_VALUE:
		if( plugin != NULL )
		{
			const auto args = inputValue.SetParamElementValueArgs;
			retval.UIntValue = plugin->SetParamElementValue( args->ParameterNumber, args->ElementNumber, args->NewParameterValue.FloatValue );
		}
		break;
	case FunctionCode::FF_GETPARAMETERUSAGE:
		retval.ParameterUsage = GetParamUsage( inputValue.UIntValue );
		break;
	case FunctionCode::FF_GETPLUGINSHORTNAME:
		retval.StrValue = GetPluginShortName();
		break;
	case FunctionCode::FF_SET_BEATINFO:
		if( plugin != NULL )
		{
			plugin->SetBeatInfo( *inputValue.BeatInfo );
			retval.UIntValue = FF_SUCCESS;
		}
		break;
	case FunctionCode::FF_SET_HOSTINFO:
		if( plugin != NULL )
		{
			const auto args = inputValue.SetHostInfoArgs;
			plugin->SetHostInfo( args->name, args->version );
			retval.UIntValue = FF_SUCCESS;
		}
		break;
	case FunctionCode::FF_SET_SAMPLERATE:
		if( plugin != NULL )
		{
			plugin->SetSampleRate( inputValue.UIntValue );
			retval.UIntValue = FF_SUCCESS;
		}
		break;
	case FunctionCode::FF_GET_RANGE:
		//TODO: ffgl@subpixels.com - investigate whether inputValue should be used to pass back the range
		retval = GetParamRange( inputValue );
		break;

		// Previously used function codes that are no longer supported:
		// case FunctionCode::FF_INITIALISE:
		/**
		 * We're dropping the old FFGL 1.6 and lower Initialise here. FFGL 2.0 removed old stuff and made support for newer stuff mandatory
		 * so hosts need a way to know they cannot use this plugin if they're dependant on the old behaviour. If the host isn't dependant on the old
		 * behaviour it will have to Update to build using the FFGL 2.0 sdk and instead invoke the initialise_v2 opcode. This way
		 * the plugin and host both agree that it's okay not to support the old behaviour.
		 */
		//case FunctionCode::FF_INSTANTIATE:
		//case FunctionCode::FF_DEINSTANTIATE:
		//case FunctionCode::FF_PROCESSFRAME:
		//case FunctionCode::FF_PROCESSFRAMECOPY:
	}

	return retval;
}

}//End namespace ffgl
